﻿using Celeste;
using Celeste.Mod;
using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.Utils;
using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace CutsceneHelper
{
    public class Cutscene : CutsceneEntity
    {
        public CutsceneTrigger cutsceneTrigger;
        public string CutscenePathName;
        public bool lockMovement;
        public bool endLevel;
        public bool active;
        CutsceneInterpreter interpreter;

        public Cutscene(string ID, bool endLvl, bool lockMovement, CutsceneTrigger trigger) : base(true, false)
        {
            CutscenePathName = ID;
            endLevel = endLvl;
            this.lockMovement = lockMovement;
            cutsceneTrigger = trigger;
            interpreter = new CutsceneInterpreter(ID, endLevel, lockMovement);
            interpreter.cutsceneTrigger = cutsceneTrigger;
            interpreter.cutscene = this;
        }

        public override void OnBegin(Level level)
        {
            Player player = Engine.Scene.Tracker.GetEntity<Player>();
            player.ForceCameraUpdate = true;
            Add(new Coroutine(CutsceneRoutine(level)));
            active = true;
        }

        public IEnumerator CutsceneRoutine(Level level)
        {
            Player player = Engine.Scene.Tracker.GetEntity<Player>();

            // lock movement
            if (lockMovement)
            {
                player.StateMachine.State = 11;
                player.StateMachine.Locked = true;
            }
            else
            {
                player.DummyFriction = true;
                player.StateMachine.Locked = false;
                //player.StateMachine.State = 0;
            }

            player.ForceCameraUpdate = true;
            //yield return Textbox.Say(dialogID);
            cutsceneTrigger.Add(new Coroutine(interpreter.InterpretScripts()));

            // EndCutscene(level, true);
            yield return null;
        }

        public override void OnEnd(Level level)
        {
            active = false;
            if (cutsceneTrigger.infinite)
            {
                cutsceneTrigger.triggered = false;
            }
            Player player = Engine.Scene.Tracker.GetEntity<Player>();
            /*
            if (interpreter.reader != null)
            {
                interpreter.reader.Dispose();
                interpreter.reader.Close();
            }
            */

            if (endLevel)
            {
                level.CompleteArea(true, false);
                player.StateMachine.State = Player.StDummy;
                RemoveSelf();
                return;
            }

            if (WasSkipped)
            {
                Logger.Log("CutsceneHelper", "Was Skipped");
                interpreter.InterpretSkipped();
                /*
                if (interpreter.reader != null)
                {
                    interpreter.reader.Dispose();
                    interpreter.reader.Close();

                } */
                //interpreter.Paused = true;
                interpreter.Interpreting = false;
            }
            player.DummyFriction = true;
            player.StateMachine.Locked = false;
            player.StateMachine.State = 0;
            player.ForceCameraUpdate = false;

            /*
            foreach (StreamReader reader in CutsceneInterpreter.textReaders)
            {
                if (reader != null)
                {
                   reader.Dispose();
                   reader.Close();
                }

            } */
        }
    }
}
