﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using Celeste;
using Celeste.Mod;
using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.Utils;
using YamlDotNet.Serialization;

namespace CutsceneHelper
{

    public class CSHelperModule : EverestModule
    {

        // Only one alive module instance can exist at any given time.
        public static CSHelperModule Instance;

        public CSHelperModule()
        {
            Instance = this;
        }
        // no save data needed
        public override Type SaveDataType => null;

        public override Type SettingsType => null;

        public override void LoadContent(bool firstLoad)
        {
            //FrostTempleModule.SpriteBank = new Monocle.SpriteBank(GFX.Game, "Graphics/FrostHelper/CustomSprites.xml");
        }

        private void resetChasers()
        {
            CutsceneInterpreter.storedChasersPos = new Vector2[] { };
            CutsceneInterpreter.storedChaserCount = 0;
            CutsceneInterpreter.storedChasersCanChangeMusic = false;
            CutsceneInterpreter.badelineChaserIndex = 0;
            forcedAim = null;
            foreach (TextReader reader in CutsceneInterpreter.textReaders)
            {
                if (reader != null)
                {
                    reader.Dispose();
                    reader.Close();
                }

            }
        }
        public static bool justSwappedCassette;
        public static int casetteIndex;
        public static Vector2? forcedAim = null;
        // Set up any hooks, event handlers and your mod in general here.
        // Load runs before Celeste itself has initialized properly.
        public override void Load()
        {
            Everest.Events.Level.OnLoadEntity += Level_OnLoadEntity;
            Everest.Events.Level.OnLoadLevel += Level_OnLoadLevel;
            Everest.Events.Level.OnTransitionTo += Level_OnTransitionTo;
            Everest.Events.Level.OnEnter += Level_OnEnter;
            Everest.Events.Level.OnExit += Level_OnExit;
            On.Celeste.Editor.MapEditor.LoadLevel += MapEditor_LoadLevel;
            On.Celeste.CassetteBlockManager.SetActiveIndex += CassetteBlockManager_SetActiveIndex;
            On.Celeste.CoreModeToggle.OnPlayer += CoreModeToggle_OnPlayer;
            CutsceneHelper.CSHelperModule.OnInterpretFunc += CutsceneHelperModule_OnInterpretFunc;
            CutsceneHelper.CSHelperModule.OnInterpretFuncSkipped += CutsceneHelperModule_OnInterpretFuncSkipped;
            On.Monocle.VirtualJoystick.Update += VirtualJoystick_Update;
        }

        private void VirtualJoystick_Update(On.Monocle.VirtualJoystick.orig_Update orig, VirtualJoystick self)
        {
            orig(self);
            if (forcedAim != null)
            {
                SetProperty<VirtualJoystick>(self, "Value", (Vector2)forcedAim);
            }
        }

        public static void SetProperty<T>(T obj, string propertyName, object newValue)
        {
            // add a check here that the object obj and propertyName string are not null
            foreach (FieldInfo fi in obj.GetType().GetFields(BindingFlags.Instance | BindingFlags.NonPublic))
            {
                if (fi.Name.ToLower().Contains(propertyName.ToLower()))
                {
                    fi.SetValue(obj, newValue);
                    break;
                }
            }
        }

        private void CutsceneHelperModule_OnInterpretFuncSkipped(string[] args)
        {
            if (args[0] == "extTest/spawnIntroCar")
            {
                Player player = Engine.Scene.Tracker.GetEntity<Player>();
                IntroCar car = new IntroCar(player.Position + new Vector2(20, 0));
                (Engine.Scene as Level).Add(car);
            }
        }

        private void CutsceneHelperModule_OnInterpretFunc(string[] args, CutsceneInterpreter interpreter)
        {
            if (args[0] == "extTest/spawnIntroCar")
            {
                Player player = Engine.Scene.Tracker.GetEntity<Player>();
                IntroCar car = new IntroCar(player.Position + new Vector2(20, 0));
                (Engine.Scene as Level).Add(car);
                interpreter.cutsceneTrigger.Add(new Coroutine(TextboxRoutine(interpreter)));
            }
        }


        private IEnumerator TextboxRoutine(CutsceneInterpreter interpreter)
        {
            interpreter.Paused = true;
            yield return Textbox.Say("CSHelper_CustomFunc", null);
            //Coroutine routine;
            //interpreter.cutsceneTrigger.Add(routine = new Coroutine(Textbox.Say("CSHelper_CustomFunc", null)));
            //while (!routine.Finished)
            //{
            //    interpreter.Paused = true;
            //    yield return null;
            //}
            interpreter.Paused = false;
        }

        public delegate void OnInterpretFuncHandler(string[] args, CutsceneInterpreter interpreter);
        /// <summary>
        /// Gets ran when you a function is interpreted in a cutscene file. args[0] is the function name
        /// </summary>
        public static event OnInterpretFuncHandler OnInterpretFunc;

        internal static void OnRunCustomFunc(string[] args, CutsceneInterpreter interpreter)
        {
            OnInterpretFuncHandler OnRunCSharpFunc = CSHelperModule.OnInterpretFunc;
            if (OnRunCSharpFunc == null)
            {
                return;
            }
            //Player player = Engine.Scene.Tracker.GetEntity<Player>();
            //player.Add(new Coroutine(OnRunCSharpFunc(args)));
            OnRunCSharpFunc(args, interpreter);
        }
            //=> OnInterpretFunc?.Invoke(args);

        
        public delegate void OnInterpretFunctionSkippedHandler(string[] args);
        /// <summary>
        /// Gets ran when a function is interpreted in a cutscene file when the cutscene is skipped. args[0] is the function name
        /// </summary>
        public static event OnInterpretFunctionSkippedHandler OnInterpretFuncSkipped;

        internal static void OnRunFuncSkipped(string[] args) => OnInterpretFuncSkipped?.Invoke(args);

        public static bool justSwappedCoreMode;

        private void CoreModeToggle_OnPlayer(On.Celeste.CoreModeToggle.orig_OnPlayer orig, CoreModeToggle self, Player player)
        {
            Session.CoreModes prevMode = self.SceneAs<Level>().Session.CoreMode;
            orig(self, player);
            if (prevMode == self.SceneAs<Level>().Session.CoreMode)
            {
                justSwappedCoreMode = true;
            }
        }

        private void CassetteBlockManager_SetActiveIndex(On.Celeste.CassetteBlockManager.orig_SetActiveIndex orig, CassetteBlockManager self, int index)
        {
            justSwappedCassette = true;
            casetteIndex = index;
            orig(self, index);
        }

        private void MapEditor_LoadLevel(On.Celeste.Editor.MapEditor.orig_LoadLevel orig, Celeste.Editor.MapEditor self, Celeste.Editor.LevelTemplate level, Vector2 at)
        {
            resetChasers();
            orig(self, level, at);
        }

        private void Level_OnExit(Level level, LevelExit exit, LevelExit.Mode mode, Session session, HiresSnow snow)
        {
            resetChasers();
        }

        private void Level_OnEnter(Session session, bool fromSaveData)
        {
            resetChasers();
        }

        private void Level_OnTransitionTo(Level level, LevelData next, Vector2 direction)
        {
            resetChasers();
        }

        private void Level_OnLoadLevel(Level level, Player.IntroTypes playerIntro, bool isFromLoader)
        {
            int index = 0;
            BadelineOldsite badeline;
            for (int i = 0; i < CutsceneInterpreter.storedChaserCount; i++)
            {
                //foreach (BadelineOldsite badeline2 in Engine.Scene.Tracker.GetEntities<BadelineOldsite>())
                //{
                //    Logger.Log("CutsceneHelper", "Found a badeline");
                    index++;
                //}
                level.Add(badeline = new BadelineOldsite(CutsceneInterpreter.storedChasersPos[i], index));
                DynData<BadelineOldsite> data = new DynData<BadelineOldsite>(badeline);
                data.Set("canChangeMusic", CutsceneInterpreter.storedChasersCanChangeMusic);
            }
        }

        private bool Level_OnLoadEntity(Level level, LevelData levelData, Vector2 offset, EntityData entityData)
        {
            if (entityData.Name == "CutsceneHelperTrigger")
            {
                level.Add(new CutsceneTrigger(entityData, offset, new EntityID(levelData.Name, entityData.ID)));
                return true;
            }
            return false;
        }

        // Optional, initialize anything after Celeste has initialized itself properly.
        public override void Initialize()
        {
        }

        // Unload the entirety of your mod's content, remove any event listeners and undo all hooks.
        public override void Unload()
        {
            Everest.Events.Level.OnLoadLevel -= Level_OnLoadLevel;
            Everest.Events.Level.OnLoadEntity -= Level_OnLoadEntity;
        }

    }
}
