﻿using Celeste;
using Celeste.Mod;
using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.Utils;
using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace CutsceneHelper
{
    public class CutsceneInterpreter : Component
    {
        public string CutscenePathName;
        //public TextReader reader;
        public bool Interpreting;
        Vector2 playerInitPos;
        bool endLevel;
        public CutsceneTrigger cutsceneTrigger;
        public static int badelineChaserIndex = 0;
        public bool lockMovement;
        Level Level;
        Dictionary<string, object> Vars = new Dictionary<string, object>()
        {

        };

        // Permament Chasers
        public static int storedChaserCount = 0;
        public static bool storedChasersCanChangeMusic = false;
        public static Vector2[] storedChasersPos = new Vector2[] { };

        public CutsceneInterpreter(string ID, bool endLvl, bool lockMovement) : base(true, false)
        {
            CutscenePathName = ID;
            endLevel = endLvl;
            this.lockMovement = lockMovement;
        }

        public static StreamReader[] textReaders = new StreamReader[] { };

        public IEnumerator InterpretScripts(string function = "", bool first = true)
        {
            //bool skipped = WasSkipped;
            Player player = Engine.Scene.Tracker.GetEntity<Player>();
            //BadelineDummy badeline = Engine.Scene.Tracker.GetEntity<BadelineDummy>();
            Level level = Engine.Scene as Level;

            while (player == null)
            {
                // player is null, halt execution
                player = Engine.Scene.Tracker.GetEntity<Player>();
                yield return null;
            }

            if (function == "")
            {
                Logger.Log("CutsceneHelper", "Reading file: " + CutscenePathName);
            } else
            {
                Logger.Log("CutsceneHelper", "Reading file: " + CutscenePathName + ", function: " + function);
            }
            
            if (Everest.Content.Map.ContainsKey("Cutscenes/" + CutscenePathName))
            {
                string[] lines = new string[0] { };
                using (StreamReader reader = new StreamReader(Everest.Content.Map["Cutscenes/" + CutscenePathName].Stream))
                {
                    
                    while(!reader.EndOfStream)
                    {
                        Array.Resize(ref lines, lines.Length + 1);
                        lines[lines.Length - 1] = reader.ReadLine();
                    }
                    reader.Dispose();
                    reader.Close();
                }
                int currentIndex = -1;
                {
                    
                    level = cutsceneTrigger.Scene as Level;
                    //Array.Resize(ref textReaders, textReaders.Length + 1);
                    //textReaders[textReaders.Length - 1] = reader;
                    //int index = textReaders.Length - 1;
                    
                    playerInitPos = player.Position;
                    bool reachedFunctionStart = false;
                    if (function == "")
                    {
                        reachedFunctionStart = true;
                    }
                    Interpreting = true;
                    while (Interpreting)
                    {
                        currentIndex++;
                        if (currentIndex >= lines.Length)
                        {
                            break;
                        }
                        while (player == null)
                        {
                            // player is null, halt execution
                            player = Engine.Scene.Tracker.GetEntity<Player>();
                            yield return null;
                        }
                        while (Paused)
                        {
                            yield return null;
                        }
                        string nextCommand = lines[currentIndex].TrimStart(' ');

                        if (!reachedFunctionStart)
                        {
                            if (nextCommand.StartsWith($"{function}:"))
                            {
                                reachedFunctionStart = true;
                                Logger.Log("CutsceneHelper", "Reached " + function);
                            }
                        } else
                        {
                            if (nextCommand == null || nextCommand == "" || nextCommand == "ifSkipped:" || nextCommand.Contains(":"))
                            {
                                if (first)
                                {
                                    Logger.Log("CutsceneHelper", "Ended reading file.");
                                    break;
                                }
                                Logger.Log("CutsceneHelper", $"Ended reading function {function}.");
                                break;
                            }
                            if (!nextCommand.StartsWith("#"))
                            {
                                string[] args = nextCommand.Split(null);

                                #region QTE
                                if (args[0] == "initDashAssist")
                                {
                                    bool lockMovement = ParseArg(args, 1, true);
                                    bool dashAssist = SaveData.Instance.Assists.DashAssist;
                                    SaveData.Instance.Assists.DashAssist = true;
                                    CustomPlayerDashAssist.DashAssistInit(lockMovement);
                                    while (Engine.DashAssistFreeze)
                                    {
                                        yield return null;
                                    }
                                    SaveData.Instance.Assists.DashAssist = dashAssist;
                                }
                                
                                #endregion
                                // say ID
                                if (nextCommand.StartsWith("say"))
                                {
                                    yield return Textbox.Say(args[1]);
                                }

                                // levelFlash color
                                if (nextCommand.StartsWith("levelFlash"))
                                {
                                    (cutsceneTrigger.Scene as Level).Flash(ParseArg(args, 1, Color.White), false);
                                }

                                // wait amt 
                                if (args[0] == "wait")
                                {
                                    yield return ParseArg(args, 1, 0f);
                                }

                                // NEW in 1.1
                                // playMusic music progress
                                if (nextCommand.StartsWith("playMusic"))
                                {
                                    string music = ParseArg(args, 1, "");
                                    // backwards compat with 1.1
                                    if (music.StartsWith("event:"))
                                    {
                                        Audio.Play(music);

                                    }
                                    else
                                    {
                                        int Progress = (int)ParseArg(args, 2, 0);
                                        Session session = (cutsceneTrigger.Scene as Level).Session;
                                        session.Audio.Music.Event = SFX.EventnameByHandle(music);
                                        bool flag = Progress != 0;
                                        if (flag)
                                        {
                                            session.Audio.Music.Progress = Progress;
                                        }
                                        session.Audio.Apply(false);
                                    }


                                    //Audio.Play(music);
                                }

                                // playSound soundEventNameByHandle
                                if (args[0] == "playSound")
                                {
                                    string music = ParseArg(args, 1, "");
                                    Audio.Play(SFX.EventnameByHandle(music));

                                }

                                // NEW in 1.3
                                // playSound soundPath xpos ypos
                                if (nextCommand.StartsWith("playSound"))
                                {
                                    string sound = ParseArg(args, 1, "");
                                    Vector2 position = new Vector2(ParseArg(args, 2, 0), ParseArg(args, 3, 0));
                                    Audio.Play(sound, position);
                                }

                                // instantWarp roomName spawnThunder
                                if (nextCommand.StartsWith("instantWarp"))
                                {
                                    level = cutsceneTrigger.Scene as Level;
                                    level.OnEndOfFrame += delegate
                                    {
                                        string destLevel = ParseArg(args, 1, "1");
                                        bool spawnThunder = ParseArg(args, 2, true);
                                        new Vector2(level.LevelOffset.X + level.Bounds.Width - player.X, player.Y - level.LevelOffset.Y);
                                        Vector2 levelOffset = level.LevelOffset;
                                        Vector2 value = player.Position - level.LevelOffset;
                                        Vector2 value2 = level.Camera.Position - level.LevelOffset;
                                        Facings facing = player.Facing;
                                        level.Remove(player);
                                        level.UnloadLevel();
                                        level.Session.Dreaming = true;
                                        level.Session.Level = destLevel;
                                        level.Session.RespawnPoint = level.GetSpawnPoint(new Vector2(level.Bounds.Left, level.Bounds.Top));
                                        level.LoadLevel(Player.IntroTypes.Transition);
                                        level.Camera.Position = level.LevelOffset + value2;
                                        level.Session.Inventory.Dashes = 1;
                                        //player.Dashes = Math.Min(player.Dashes, 1);
                                        level.Add(player);
                                        player.Position = level.LevelOffset + value;
                                        player.Facing = facing;
                                        player.Hair.MoveHairBy(level.LevelOffset - levelOffset);
                                        if (level.Wipe != null)
                                        {
                                            level.Wipe.Cancel();
                                        }
                                        level.Flash(Color.White);
                                        level.Shake();
                                        if (spawnThunder)
                                        {
                                            level.Add(new LightningStrike(new Vector2(player.X + 60f, level.Bounds.Bottom - 180), 10, 200f));
                                            level.Add(new LightningStrike(new Vector2(player.X + 220f, level.Bounds.Bottom - 180), 40, 200f, 0.25f));
                                            Audio.Play("event:/new_content/game/10_farewell/lightning_strike");
                                        }
                                    };
                                }
                                #region Player

                                // reloadPlayerPos
                                // Reloads the player's position from before the cutscene
                                if (nextCommand.StartsWith("reloadPlayerPos"))
                                {
                                    player.Position = playerInitPos;
                                }

                                // playerDash left/right/none up/down/none
                                if (nextCommand.StartsWith("playerDash"))
                                {
                                    player.StateMachine.Locked = false;
                                    player.DummyFriction = true;
                                    string dirStr = ParseArg(args, 1, "none");
                                    int dir = dirStr == "right" ? 1 : -1;
                                    if (dirStr == "none") dir = 0;
                                    string dir2Str = ParseArg(args, 2, "none");
                                    int dir2 = dir2Str == "down" ? 1 : -1;
                                    if (dir2Str == "none") dir2 = 0;
                                    player.OverrideDashDirection = new Vector2?(new Vector2(dir, dir2));
                                    player.StateMachine.State = player.StartDash();
                                    yield return ParseArg(args, 3, 0.6f);
                                    player.OverrideDashDirection = null;
                                    if (player.StateMachine.State != Player.StBoost)
                                        player.StateMachine.State = 11;
                                    player.StateMachine.Locked = true;
                                }

                                // playerSetAim left/right/none up/down/none time
                                if (args[0] == "playerSetAim" || args[0] == "setPlayerAim")
                                {
                                    bool locked = player.StateMachine.Locked;
                                    player.StateMachine.Locked = false;
                                    //player.DummyFriction = true;
                                    string dirStr = ParseArg(args, 1, "none");
                                    int dir = dirStr == "right" ? 1 : -1;
                                    if (dirStr == "none") dir = 0;
                                    string dir2Str = ParseArg(args, 2, "none");
                                    int dir2 = dir2Str == "down" ? 1 : -1;
                                    if (dir2Str == "none") dir2 = 0;
                                    CSHelperModule.forcedAim = new Vector2(dir, dir2);
                                    yield return ParseArg(args, 3, 1f);
                                    CSHelperModule.forcedAim = null;
                                    //if (player.StateMachine.State != Player.StBoost)
                                    while (player.StateMachine.State == Player.StBoost)
                                    {
                                        yield return null;
                                    }
                                    player.StateMachine.State = Player.StDummy;
                                    
                                    player.StateMachine.Locked = true; // locked
                                }

                                // playerWalk x speed
                                if (nextCommand.StartsWith("playerWalk "))
                                {
                                    float amt = ParseArg(args, 1, 0);
                                    float speed = ParseArg(args, 2, 1);
                                    yield return player.DummyWalkTo(player.Position.X + amt, false, speed, false);
                                }

                                // playerWalkTo x speed
                                if (nextCommand.StartsWith("playerWalkTo"))
                                {
                                    float pos = ParseArg(args, 1, 0);
                                    float speed = ParseArg(args, 2, 1);
                                    pos += (cutsceneTrigger.Scene as Level).LevelOffset.X;
                                    yield return player.DummyWalkTo(pos, false, speed, false);
                                }

                                // setPlayerSpeedX xspeed
                                if (nextCommand.StartsWith("setPlayerSpeedX"))
                                {
                                    float xspeed = ParseArg(args, 1, 0);
                                    player.Speed.X = xspeed;
                                }

                                // setPlayerSpeedY yspeed
                                if (nextCommand.StartsWith("setPlayerSpeedY"))
                                {
                                    float yspeed = ParseArg(args, 1, 0);
                                    player.Speed.Y = yspeed;
                                }

                                // setPlayerPos x y relativeToPlayer?
                                if (nextCommand.StartsWith("setPlayerPos"))
                                {
                                    float x = ParseArg(args, 1, 0);
                                    float y = ParseArg(args, 2, 0);
                                    bool relative = ParseArg(args, 3, false);
                                    if (relative)
                                    {
                                        x += player.Position.X;
                                        y += player.Position.Y;
                                    } else
                                    {
                                        x += (cutsceneTrigger.Scene as Level).LevelOffset.X;
                                        y += (cutsceneTrigger.Scene as Level).LevelOffset.Y;
                                    }
                                    player.Position.X = x;
                                    player.Position.Y = y;
                                }

                                // playerJump xspeed direction waitTillGrounded? setDirection?
                                if (nextCommand.StartsWith("playerJump"))
                                {
                                    float xspeed = ParseArg(args, 1, 0);
                                    string dirStr = ParseArg(args, 2, "right");
                                    bool waitTillGrounded = ParseArg(args, 3, false);
                                    bool setDirection = ParseArg(args, 4, false);
                                    int dir = dirStr == "right" ? 1 : -1;
                                    PlayerJump(dir, xspeed, setDirection);
                                    if (waitTillGrounded)
                                    {
                                        yield return WaitForPlayerGrounded();
                                        player.Speed.X = 0;
                                    }
                                }

                                // waitTillPlayerGrounded
                                if (nextCommand.StartsWith("waitTillPlayerGrounded"))
                                {
                                    yield return WaitForPlayerGrounded();
                                }

                                // playerSpritePlay
                                if (nextCommand.StartsWith("playerSpritePlay"))
                                {
                                    string anim = args[1];
                                    player.DummyAutoAnimate = false;
                                    player.Sprite.Play(anim, false, false);
                                }

                                // playerClimbWall direction
                                if (nextCommand.StartsWith("playerClimbWall"))
                                {
                                    string dirStr = ParseArg(args, 1, "right");
                                    int dir = dirStr == "right" ? 1 : -1;
                                    player.DummyAutoAnimate = false;
                                    player.Sprite.Play("climbUp", false, false);
                                    player.Facing = (Facings)dir;
                                    while (player.CollideCheck<Solid>(player.Position + new Vector2(dir, 0f)))
                                    {
                                        player.Speed.X = 64f;
                                        player.Speed.Y = Math.Min(player.Speed.Y, -45f);
                                        yield return null;
                                    }
                                    player.DummyAutoAnimate = true;
                                }

                                // setPlayerDashes count
                                if (nextCommand.StartsWith("setPlayerDashes"))
                                {
                                    int dashes = (int)ParseArg(args, 1, 0);
                                    player.Dashes = dashes;
                                    player.UpdateHair(true);
                                }

                                // setPlayerMaxDashes count
                                if (nextCommand.StartsWith("setPlayerMaxDashes"))
                                {
                                    int dashes = (int)ParseArg(args, 1, 0);
                                    (cutsceneTrigger.Scene as Level).Session.Inventory.Dashes = dashes;
                                    if (player.Dashes > dashes)
                                    {
                                        player.Dashes = dashes;
                                    }
                                    player.UpdateHair(true);
                                }

                                // playerRefilOnGround true?
                                if (nextCommand.StartsWith("playerRefilOnGround") || nextCommand.StartsWith("playerRefillOnGround"))
                                {
                                    bool playerRefilOnGround = ParseArg(args, 1, true);
                                    (cutsceneTrigger.Scene as Level).Session.Inventory.NoRefills = !playerRefilOnGround;
                                }
                                #endregion

                                #region Camera

                                // cameraTo targetX targetY duration
                                if (nextCommand.StartsWith("cameraTo"))
                                {
                                    float xtarget = ParseArg(args, 1, 0);
                                    float ytarget = ParseArg(args, 2, 0);
                                    float duration = ParseArg(args, 3, 0);
                                    yield return CutsceneEntity.CameraTo(new Vector2(xtarget, ytarget), duration);
                                }

                                // cameraToPlayer duration
                                if (nextCommand.StartsWith("cameraToPlayer"))
                                {
                                    float duration = ParseArg(args, 1, 0);
                                    yield return CutsceneEntity.CameraTo(player.Position, duration);
                                    //(cutsceneTrigger.Scene as Level).Camera.Approach(player.Position, duration);
                                }

                                // levelShake duration
                                if (nextCommand.StartsWith("levelShake"))
                                {
                                    float duration = ParseArg(args, 1, 0);
                                    (cutsceneTrigger.Scene as Level).Shake(duration);
                                }

                                // cameraZoom zoom
                                if (nextCommand.StartsWith("cameraZoom"))
                                {
                                    
                                    float zoom = ParseArg(args, 1, 0);
                                    (cutsceneTrigger.Scene as Level).Camera.Zoom = zoom;
                                }
                                #endregion

                                // crumbleWallsOnRumble
                                if (nextCommand.StartsWith("crumbleWallsOnRumble"))
                                {
                                    foreach (Entity entity in (cutsceneTrigger.Scene as Level).Tracker.GetEntities<CrumbleWallOnRumble>())
                                    {
                                        CrumbleWallOnRumble crumbleWallOnRumble = (CrumbleWallOnRumble)entity;
                                        crumbleWallOnRumble.Break();
                                    }
                                }

                                // swapCoreMode
                                if (nextCommand.StartsWith("swapCoreMode"))
                                {
                                    string arg = "cold";
                                    if ((cutsceneTrigger.Scene as Level).CoreMode == Session.CoreModes.Cold)
                                    {
                                        arg = "hot";
                                    }
                                    changeCoreMode(arg);
                                }

                                // setCoreMode hot/cold
                                if (nextCommand.StartsWith("setCoreMode"))
                                {
                                    string arg = ParseArg(args, 1, "hot");
                                    changeCoreMode(arg);
                                    /*
                                    if (arg == "hot")
                                    {
                                        yield return (cutsceneTrigger.Scene as Level).CoreMode = Session.CoreModes.Hot;
                                    }
                                    else if (arg == "cold")
                                    {
                                        yield return (cutsceneTrigger.Scene as Level).CoreMode = Session.CoreModes.Cold;
                                    } else
                                    {
                                        yield return (cutsceneTrigger.Scene as Level).CoreMode = Session.CoreModes.None;
                                    }
                                    */
                                }

                                #region Badeline

                                // badelineAppear x y relativeToPlayer? silent? retArg:BadelineDummyVarName
                                if (nextCommand.StartsWith("badelineAppear"))
                                {
                                    float x = ParseArg(args, 1, 0);
                                    float y = ParseArg(args, 2, 0);
                                    bool relativeToPlayer = ParseArg(args, 3, false);
                                    bool silent = ParseArg(args, 4, false);
                                    string retVarName = ParseArg(args, 5, "");
                                    BadelineDummy badeline;
                                    if (relativeToPlayer)
                                    {
                                        (cutsceneTrigger.Scene as Level).Add(badeline = new BadelineDummy(player.Position + new Vector2(x, y)));
                                    } else
                                    {
                                        (cutsceneTrigger.Scene as Level).Add(badeline = new BadelineDummy((cutsceneTrigger.Scene as Level).LevelOffset + new Vector2(x, y)));
                                        //(cutsceneTrigger.Scene as Level).Add(badeline = new BadelineDummy(new Vector2(x, y)));
                                    }
                                    badeline.Floatness = 4f;
                                    badeline.Appear((cutsceneTrigger.Scene as Level), silent);
                                    Vars.Add(retVarName, badeline);
                                    //badelineAutoFloat = true;
                                    //badelineAutoWalk = false;
                                    //badelineWalkApproach = 0f;
                                }

                                // badelineVanish badelineArg
                                if (nextCommand.StartsWith("badelineVanish"))
                                {
                                    string badelineVarName = ParseArg(args, 1, "");
                                    BadelineDummy badeline = Vars[badelineVarName] as BadelineDummy;
                                    if (badeline != null)
                                        badeline.Vanish();
                                    Vars.Remove(badelineVarName);
                                }

                                // badelineFloat x y badelineArg threaded?
                                if (nextCommand.StartsWith("badelineFloat"))
                                {
                                    float posx = ParseArg(args, 1, 0);
                                    float posy = ParseArg(args, 2, 0);
                                    string badelineVarName = ParseArg(args, 3, "");
                                    bool threaded = ParseArg(args, 4, false);
                                    BadelineDummy badeline; // = Vars[badelineVarName] as BadelineDummy;
                                    if (!Vars.ContainsKey(badelineVarName))
                                    {
                                        badeline = null;
                                    }
                                    else
                                    {
                                        badeline = Vars[badelineVarName] as BadelineDummy;
                                    }
                                    if (badeline != null)
                                    {
                                        if (threaded)
                                        {
                                            cutsceneTrigger.Add(new Coroutine(badelineFloat(badeline, posx, posy)));
                                        } else
                                        {
                                            yield return badeline.FloatTo(badeline.Position + new Vector2(posx, posy));
                                        }
                                        
                                    }
                                }

                                // badelineSurround x y badelineArg threaded?
                                if (nextCommand.StartsWith("badelineSurround"))
                                {
                                    float posx = ParseArg(args, 1, 0);
                                    float posy = ParseArg(args, 2, 0);
                                    string badelineVarName = ParseArg(args, 3, "");
                                    bool threaded = ParseArg(args, 4, false);
                                    BadelineDummy badeline;
                                    if (!Vars.ContainsKey(badelineVarName))
                                    {
                                        badeline = null;
                                    } else
                                    {
                                        badeline = Vars[badelineVarName] as BadelineDummy;
                                    }
                                    
                                    //if (badeline != null)
                                    {
                                        if (threaded)
                                        {
                                            if (badeline != null)
                                            {
                                                cutsceneTrigger.Add(new Coroutine(BadelineAround(badeline.Position, badeline.Position + new Vector2(posx, posy), badelineVarName, badeline)));
                                            } else
                                            {
                                                cutsceneTrigger.Add(new Coroutine(BadelineAround(player.Position, player.Position + new Vector2(posx, posy), badelineVarName)));
                                            }
                                                
                                        }
                                        else
                                        {
                                            if (badeline != null)
                                            {
                                                yield return BadelineAround(badeline.Position, badeline.Position + new Vector2(posx, posy), badelineVarName, badeline);
                                            } else
                                            {
                                                yield return BadelineAround(player.Position, player.Position + new Vector2(posx, posy), badelineVarName);
                                            }

                                        }

                                    }
                                }

                                // NEW in 1.1
                                // badelineSpritePlay badeline anim
                                if (nextCommand.StartsWith("badelineSpritePlay"))
                                {
                                    string badelineVarName = ParseArg(args, 1, "");
                                    string anim = ParseArg(args, 2, "");
                                    BadelineDummy badeline; // = Vars[badelineVarName] as BadelineDummy;
                                    if (!Vars.ContainsKey(badelineVarName))
                                    {
                                        badeline = null;
                                    }
                                    else
                                    {

                                        badeline = Vars[badelineVarName] as BadelineDummy;
                                        badeline.Sprite.Play(anim);
                                    }
                                }

                                // NEW in 1.1
                                // spawnBadelineChaser x y relativeToPlayer? canChangeMusic? permament?
                                if (nextCommand.StartsWith("spawnBadelineChaser"))
                                {
                                    
                                    float x = ParseArg(args, 1, 0);
                                    float y = ParseArg(args, 2, 0);
                                    bool relativeToPlayer = ParseArg(args, 3, false);
                                    bool canChangeMusic = ParseArg(args, 4, false);
                                    bool permament = ParseArg(args, 5, false);
                                    //foreach (BadelineOldsite badeline2 in Scene.Tracker.GetEntities<BadelineOldsite>())
                                   // {
                                    badelineChaserIndex++;
                                    //}
                                    Vector2 pos = new Vector2(x, y);
                                    if (relativeToPlayer)
                                    {
                                        pos += player.Position;
                                    }
                                    else
                                    {
                                        pos += (cutsceneTrigger.Scene as Level).LevelOffset;
                                    }
                                    BadelineOldsite badeline;
                                    (cutsceneTrigger.Scene as Level).Add(badeline = new BadelineOldsite(pos, badelineChaserIndex));
                                    DynData<BadelineOldsite> data = new DynData<BadelineOldsite>(badeline);
                                    data.Set("canChangeMusic", canChangeMusic);
                                    if (permament)
                                    {
                                        storedChaserCount++;
                                        //storedChasersPos.SetValue(pos, storedChaserCount - 1);
                                        Array.Resize<Vector2>(ref storedChasersPos, storedChasersPos.Length+1);
                                        storedChasersPos[storedChasersPos.Length-1] = pos;
                                        if (canChangeMusic)
                                        {
                                            storedChasersCanChangeMusic = true;
                                        }
                                    }
                                    //Vars.Add(retVarName, badeline);
                                }

                                // NEW in 1.1
                                // badelineChasersStopChasing
                                if (nextCommand.StartsWith("badelineChasersStopChasing"))
                                {
                                    bool threaded = true; //ParseArg(args, 1, false);
                                    foreach (BadelineOldsite badeline in cutsceneTrigger.Scene.Tracker.GetEntities<BadelineOldsite>())
                                    {
                                        if (threaded)
                                        {
                                            cutsceneTrigger.Add(new Coroutine(badeline_StopChasing(badeline), true));
                                        }
                                    }
                                    badelineChaserIndex = 0;
                                }
                                #endregion

                                #region flowControl
                                if (nextCommand.StartsWith("exec "))
                                {
                                    string calledFunction = ParseArg(args, 1, "");
                                    cutsceneTrigger.Add(new Coroutine(InterpretScripts(calledFunction, false)));
                                }

                                // execWait functionName
                                if (nextCommand.StartsWith("execWait"))
                                {
                                    string calledFunction = ParseArg(args, 1, "");
                                    Logger.Log("Cutscene Helper", "Calling a function:");
                                    yield return InterpretScripts(calledFunction, false);
                                }

                                // if varName condition value
                                if (nextCommand.StartsWith("if "))
                                {
                                    string varName = ParseArg(args, 1, "");
                                    string varValue = Vars[varName].ToString();
                                    string condition = ParseArg(args, 2, "=");
                                    string targetValue = ParseArg(args, 3, "");
                                    if (condition == "=")
                                    {
                                        if (varValue == targetValue)
                                        {
                                            OnTrue();
                                        } else
                                        {
                                            OnFalse();
                                        }
                                    }
                                    if (condition == ">")
                                    {
                                        if (float.Parse(varValue) > float.Parse(targetValue))
                                        {
                                            OnTrue();
                                        }
                                        else
                                        {
                                            OnFalse();
                                        }
                                    }
                                    if (condition == "<")
                                    {
                                        if (float.Parse(varValue) < float.Parse(targetValue))
                                        {
                                            OnTrue();
                                        }
                                        else
                                        {
                                            OnFalse();
                                        }
                                    }
                                    if (condition == ">=")
                                    {
                                        if (float.Parse(varValue) >= float.Parse(targetValue))
                                        {
                                            OnTrue();
                                        }
                                        else
                                        {
                                            OnFalse();
                                        }
                                    }
                                    if (condition == "<=")
                                    {
                                        if (float.Parse(varValue) <= float.Parse(targetValue))
                                        {
                                            OnTrue();
                                        }
                                        else
                                        {
                                            OnFalse();
                                        }
                                    }

                                    void OnTrue()
                                    {

                                    }

                                    void OnFalse()
                                    {
                                        int iflevel = 1;
                                        for (; ; )
                                        {
                                            currentIndex++;
                                            string next = lines[currentIndex].TrimStart(' ');
                                            if (next.StartsWith("endIf"))
                                            {
                                                iflevel--;
                                                if (iflevel < 1)
                                                    break;
                                            }
                                            if (next.StartsWith("if"))
                                            {
                                                iflevel++;
                                            }
                                        }
                                    }
                                }

                                // label Name
                                // No actual code to this

                                // goto labelName
                                if (nextCommand.StartsWith("goto"))
                                {
                                    string labelName = ParseArg(args, 1, "");
                                    if (labelName != "")
                                    {
                                        int labelIndex = SeekLabel(labelName, lines);
                                        if (labelIndex > -1)
                                        {
                                            //Logger.Log("CutsceneHelper", reader.BaseStream.Position.ToString());

                                            //reader.BaseStream.Seek(labelIndex, SeekOrigin.Begin);
                                            //reader.DiscardBufferedData();
                                            //reader.ReadLine();
                                            currentIndex = labelIndex;
                                            //Logger.Log("CutsceneHelper", reader.ReadLine());
                                        }
                                    }
                                    
                                }

                                // waitForCassetteBlockSwap index
                                if (args[0] == "waitForCasseteBlockSwap" || args[0] == "waitForCassetteBlockSwap")
                                {
                                    int targetIndex = (int)ParseArg(args, 1, -1);
                                    CSHelperModule.justSwappedCassette = false;
                                    for (; ; )
                                    {
                                        if (!CSHelperModule.justSwappedCassette)
                                        {
                                            yield return null;
                                        } else
                                        {
                                            if (targetIndex == -1 || CSHelperModule.casetteIndex == targetIndex)
                                            {
                                                break;
                                            }
                                            CSHelperModule.justSwappedCassette = false;
                                        }
                                        
                                    }
                                    
                                }

                                // waitForCoreModeSwap hot/cold
                                if (nextCommand.StartsWith("waitForCoreModeSwap"))
                                {
                                    //int targetIndex = (int)ParseArg(args, 1, -1);
                                    string arg = ParseArg(args, 1, "");
                                    CSHelperModule.justSwappedCoreMode = false;
                                    for (; ; )
                                    {
                                        if (!CSHelperModule.justSwappedCoreMode)
                                        {
                                            yield return null;
                                        }
                                        else
                                        {
                                            string value = (cutsceneTrigger.Scene as Level).CoreMode == Session.CoreModes.Hot ? "hot" : "cold";
                                            if (arg == "" || arg == value)
                                            {
                                                break;
                                            }
                                            CSHelperModule.justSwappedCoreMode = false;
                                        }

                                    }

                                }

                                #endregion

                                #region Custom

                                // ext name
                                // Runs a function from an external mod
                                //if (args[0] == "ext")
                                {
                                    ExecCustomFunc(args);
                                }
                                #endregion

                                #region Debug
                                // debugLog text
                                if (nextCommand.StartsWith("debugLog"))
                                {
                                    Logger.Log("CutsceneHelper", nextCommand.Substring("debugLog ".Length));
                                }
                                #endregion

                                #region Variables
                                // setVar varName value
                                if (nextCommand.StartsWith("setVar"))
                                {
                                    string varName = ParseArg(args, 1, "");
                                    string value = ParseArg(args, 2, "");
                                    if (!Vars.ContainsKey(varName))
                                    {
                                        Vars.Add(varName, value);
                                    } else
                                    {
                                        Vars[varName] = value;
                                    }
                                    
                                }

                                // addVarNum varName value
                                if (nextCommand.StartsWith("addVarNum"))
                                {
                                    string varName = ParseArg(args, 1, "");
                                    float value = ParseArg(args, 2, 0f);
                                    if (Vars.ContainsKey(varName))
                                    {
                                        Vars[varName] = float.Parse(Vars[varName].ToString()) + value;
                                    }
                                }

                                // getCassetteBlockIndex retVar
                                if (nextCommand.StartsWith("getCassetteBlockIndex"))
                                {
                                    string varName = ParseArg(args, 1, "");
                                    if (!Vars.ContainsKey(varName))
                                    {
                                        Vars.Add(varName, CSHelperModule.casetteIndex);
                                    }
                                    else
                                    {
                                        Vars[varName] = CSHelperModule.casetteIndex;
                                    }
                                }

                                // getCoreMode retVar
                                if (nextCommand.StartsWith("getCoreMode"))
                                {
                                    string varName = ParseArg(args, 1, "");
                                    string value = (cutsceneTrigger.Scene as Level).CoreMode == Session.CoreModes.Hot ? "hot" : "cold";
                                    if (!Vars.ContainsKey(varName))
                                    {
                                        Vars.Add(varName, value);
                                    }
                                    else
                                    {
                                        Vars[varName] = value;
                                    }
                                }

                                #endregion

                                #region Misc
                                // setCassetteBlockTempo tempo
                                if (nextCommand.StartsWith("setCassetteBlockTempo"))
                                {
                                    float Tempo = ParseArg(args, 1, 1f);
                                    SetCassetteTempo(Tempo);
                                }

                                // activateDreamBlocks threaded?
                                if (args[0] == "activateDreamBlocks")
                                {
                                    bool threaded = ParseArg(args, 1, false);
                                    level.Session.Inventory.DreamDash = true;
                                    foreach (DreamBlock dreamBlock in cutsceneTrigger.Scene.Tracker.GetEntities<DreamBlock>())
                                    {
                                        if (threaded)
                                        {
                                            cutsceneTrigger.Add(new Coroutine(dreamBlock.Activate()));
                                        } else
                                        {
                                            yield return dreamBlock.Activate();
                                        }
                                        
                                    }
                                }
                                #endregion

                                ExecSharedFuncs(args);
                            }
                        }
                        
                    }
                    
                }
            }
            if (first)
            {
                Interpreting = false;
                if (cutscene != null)
                    cutscene.EndCutscene(level, true);
            }
            if (cutsceneTrigger.infinite)
            {
                cutsceneTrigger.triggered = false;
            }

        }

        private void ExecCustomFunc(string[] args)
        {
            //Paused = true;
            CSHelperModule.OnRunCustomFunc(args, this);
            //Paused = false;
        }

        public bool Paused;

        public void changeCoreMode(string mode)
        {
            Session.CoreModes prevMode = (cutsceneTrigger.Scene as Level).CoreMode;
            if (mode == "hot")
            {
                (cutsceneTrigger.Scene as Level).CoreMode = Session.CoreModes.Hot;
            }
            else if (mode == "cold")
            {
                (cutsceneTrigger.Scene as Level).CoreMode = Session.CoreModes.Cold;
            }
            else
            {
                (cutsceneTrigger.Scene as Level).CoreMode = Session.CoreModes.None;
            }
            if (prevMode != (cutsceneTrigger.Scene as Level).CoreMode)
            {
                foreach (CoreModeListener listener in cutsceneTrigger.Scene.Tracker.GetComponents<CoreModeListener>())
                {
                    listener.OnChange?.Invoke((cutsceneTrigger.Scene as Level).CoreMode);
                }
            }
        }

        public Cutscene cutscene;
        public void SetCassetteTempo(float Tempo)
        {
            // let me speak to your manager
            CassetteBlockManager manager = cutsceneTrigger.Scene.Tracker.GetEntity<CassetteBlockManager>();
            DynData<CassetteBlockManager> data = new DynData<CassetteBlockManager>(manager);
            data.Set("tempoMult", Tempo);
            cutsceneTrigger.SceneAs<Level>().CassetteBlockTempo = Tempo;
        }

        public void ExecSharedFuncs(string[] args)
        {
            switch (args[0])
            {
                case "changeAreaData":
                    //string name = args[1];
                    //string value = args[2];
                    //FieldInfo info = typeof(AreaData).GetField(name);
                    //DynData<AreaData> data = new DynData<AreaData>(cutsceneTrigger.SceneAs<Level>().Session.MapData.Data);
                    //Type type = data.Get<object>(name).GetType();
                    //Type type = info.FieldType;
                    //type.
                    //typeof() Value
                    //dynamic value2 = Convert.ChangeType(value, type);
                    //data.Set(name, value2);
                    //info.SetValue(, value2);
                    break;
            }
        }

        public T Convert1<T>(object input)
        {
            return (T)Convert.ChangeType(input, typeof(T));
        }


        private int SeekLabel(string labelName, string[] lines)
        {
            int labelIndex = -1;
            /*
            using (StreamReader streamReader = new StreamReader(Everest.Content.Map["Cutscenes/" + CutscenePathName].Stream))
            {
                for (; ; )
                {
                    string nextLine = streamReader.ReadLine();
                    labelIndex += nextLine.Length;
                    if (nextLine.StartsWith($"label {labelName}"))
                    {
                        //labelIndex = (int)streamReader.BaseStream.Position;
                        //Logger.Log("CutsceneHelper", $"Found label {labelName} at {labelIndex}");
                        break;
                    }
                }
                streamReader.Dispose();
                streamReader.Close();
            } */
            for(int i = 0; i < lines.Length; i++)
            {
                if (lines[i].TrimStart(' ').StartsWith($"label {labelName}"))
                {
                    labelIndex = i;
                }
            }
            return labelIndex;
        }

        private IEnumerator badeline_StopChasing(BadelineOldsite badeline)
        {
            Level level = (cutsceneTrigger.Scene as Level);
            DynData<BadelineOldsite> data = new DynData<BadelineOldsite>(badeline);
            //badeline.following = false;
            //badeline.ignorePlayerAnim = true;
            data.Set("following", false);
            data.Set("ignorePlayerAnim", true);
            badeline.Sprite.Play("laugh", false, false);
            badeline.Sprite.Scale.X = 1f;
            yield return 1f;
            Audio.Play("event:/char/badeline/disappear", badeline.Position);
            level.Displacement.AddBurst(badeline.Center, 0.5f, 24f, 96f, 0.4f, null, null);
            level.Particles.Emit(BadelineOldsite.P_Vanish, 12, badeline.Center, Vector2.One * 6f);
            badeline.RemoveSelf();
            yield break;
        }

        private IEnumerator BadelineAround(Vector2 start, Vector2 around, string badelineVarName, BadelineDummy badeline = null)
        {
            Player player = Engine.Scene.Tracker.GetEntity<Player>();
            Scene Scene = Engine.Scene;
            bool removeAtEnd = badeline == null;
            bool flag = badeline == null;
            if (flag)
            {
                Scene.Add(badeline = new BadelineDummy(start));
                Vars.Add(badelineVarName, badeline);
            }
            badeline.Sprite.Play("fallSlow", false, false);
            float angle = Calc.Angle(around, start);
            float dist = (around - start).Length();
            float duration = 3f;
            for (float p = 0f; p < 1f; p += Engine.DeltaTime / duration)
            {
                float a = p * 2f;
                badeline.Position = around + Calc.AngleToVector(angle - a * 6.28318548f, dist + Calc.YoYo(p) * 16f + (float)Math.Sin((double)(p * 6.28318548f * 4f)) * 5f);
                badeline.Sprite.Scale.X = Math.Sign(around.X - badeline.X);
                bool flag2 = !removeAtEnd;
                if (flag2)
                {
                    player.Facing = (Facings)Math.Sign(badeline.X - player.X);
                }
                bool flag3 = Scene.OnInterval(0.1f);
                if (flag3)
                {
                    TrailManager.Add(badeline, Player.NormalHairColor, 1f, false, false);
                }
                yield return null;
            }
            bool flag4 = removeAtEnd;
            if (flag4)
            {
                badeline.Vanish();
                Vars.Remove(badelineVarName);
            }
            else
            {
                badeline.Sprite.Play("laugh", false, false);
            }
            yield break;
        }

        public IEnumerator WaitForPlayerGrounded()
        {
            Player player = Engine.Scene.Tracker.GetEntity<Player>();
            while (!player.OnGround(1))
            {
                yield return 0.1f;
            }
            player.DummyFriction = true;
        }

        private void PlayerJump(int direction, float xspeed, bool setDirection)
        {
            Player player = Engine.Scene.Tracker.GetEntity<Player>();
            if (setDirection)
            player.Facing = (Facings)direction;
            player.DummyFriction = false;
            player.DummyAutoAnimate = true;
            player.Speed.X = (direction * xspeed);
            player.Jump(true, true);
            player.AutoJump = true;
            player.AutoJumpTimer = 2f;
        }
        public IEnumerator badelineFloat(BadelineDummy badeline, float posx, float posy)
        {
            yield return badeline.FloatTo(badeline.Position + new Vector2(posx, posy));
        }

        public Color ParseArg(string[] args, int index, Color defaultValue)
        {
            if (args.Length > index)
            {
                //return Calc.HexToColor(ParseArg(args, index, 0f).ToString());
                if (args[index].StartsWith(">"))
                {
                    string var = args[index].Substring(1);
                    return Calc.HexToColor(Vars[var].ToString());
                } else
                {
                    return Calc.HexToColor(args[index]);
                }
                
            }
            else
            {
                return defaultValue;
            }
        }

        public float ParseArg(string[] args, int index, float defaultValue = 0f)
        {
            if (args.Length > index)
            {
                if (args[index].StartsWith(">"))
                {
                    string var = args[index].Substring(1);
                    return float.Parse(Vars[var].ToString());
                } else
                {
                    return float.Parse(args[index]);
                }  
            } else
            {
                return defaultValue;
            }
        }

        public bool ParseArg(string[] args, int index, bool defaultValue = false)
        {
            if (args.Length > index)
            {
                return bool.Parse(args[index]);
            }
            else
            {
                return defaultValue;
            }
        }

        public string ParseArg(string[] args, int index, string defaultValue = "")
        {
            if (args.Length > index)
            {
                if (args[index].StartsWith(">"))
                {
                    string var = args[index].Substring(1);
                    return Vars[var].ToString();
                } else
                {
                    return args[index].ToString();
                }
                
            }
            else
            {
                return defaultValue;
            }
        }

        


        

        public void InterpretSkipped(string function = "ifSkipped", bool first = true)
        {
            bool skipped = true;
            Player player = Engine.Scene.Tracker.GetEntity<Player>();
            Level level = Engine.Scene as Level;
            Logger.Log("CutsceneHelper", "Reading file: " + CutscenePathName + ", function: " + function);
            if (Everest.Content.Map.ContainsKey("Cutscenes/" + CutscenePathName))
            {
                string[] lines = new string[0] { };
                using (StreamReader reader = new StreamReader(Everest.Content.Map["Cutscenes/" + CutscenePathName].Stream))
                {
                    while (!reader.EndOfStream)
                    {
                        Array.Resize(ref lines, lines.Length + 1);
                        lines[lines.Length - 1] = reader.ReadLine();
                    }
                    reader.Dispose();
                    reader.Close();
                }
                {
                    int currentIndex = -1;
                    bool reachedIfSkipped = false;
                    for (; ; )
                    {
                    Start:
                        currentIndex++;
                        if (currentIndex >= lines.Length)
                        {
                            break;
                        }

                        string nextCommand = lines[currentIndex].TrimStart(' ');

                        if (!reachedIfSkipped)
                        {
                            if (nextCommand == function + ":")
                            {
                                reachedIfSkipped = true;
                                Logger.Log("CutsceneHelper", "Found function: " + function);
                            }
                            goto Start;
                        }

                        if (nextCommand == null || nextCommand == "" || nextCommand.Contains(":"))
                        {
                            if (first)
                                Logger.Log("CutsceneHelper", "Ended reading file.");
                            break;
                        }

                        if (!nextCommand.StartsWith("#"))
                            {
                                string[] args = nextCommand.Split(null);

                            #region Custom

                            // ext name
                            // Runs a function from an external mod
                            CSHelperModule.OnRunFuncSkipped(args);
                            #endregion

                            // playerWalk
                            if (nextCommand.StartsWith("playerWalk"))
                            {
                                float amt = ParseArg(args, 1, 0);
                                player.Position.X += amt;
                            }

                            // swapCoreMode
                            if (nextCommand.StartsWith("swapCoreMode"))
                            {
                                if ((cutsceneTrigger.Scene as Level).CoreMode == Session.CoreModes.Cold)
                                {
                                    (cutsceneTrigger.Scene as Level).CoreMode = Session.CoreModes.Hot;
                                }
                                else
                                {
                                    (cutsceneTrigger.Scene as Level).CoreMode = Session.CoreModes.Cold;
                                }
                            }

                            // setCoreMode hot/cold
                            if (nextCommand.StartsWith("setCoreMode"))
                            {
                                if (args[1] == "hot")
                                {
                                    (cutsceneTrigger.Scene as Level).CoreMode = Session.CoreModes.Hot;
                                }
                                else if (args[1] == "cold")
                                {
                                    (cutsceneTrigger.Scene as Level).CoreMode = Session.CoreModes.Cold;
                                }
                                else
                                {
                                    (cutsceneTrigger.Scene as Level).CoreMode = Session.CoreModes.None;
                                }
                            }

                            // reloadPlayerPos
                            // Reloads the player's position from before the cutscene
                            if (nextCommand.StartsWith("reloadPlayerPos"))
                            {
                                player.Position = playerInitPos;
                            }

                            // setPlayerPos x y relativeToPlayer?
                            if (nextCommand.StartsWith("setPlayerPos"))
                            {
                                float x = ParseArg(args, 1, 0);
                                float y = ParseArg(args, 2, 0);
                                bool relative = ParseArg(args, 3, false);
                                if (relative)
                                {
                                    x += player.Position.X;
                                    y += player.Position.Y;
                                }
                                else
                                {
                                    x += (cutsceneTrigger.Scene as Level).LevelOffset.X;
                                    y += (cutsceneTrigger.Scene as Level).LevelOffset.Y;
                                }
                                player.Position.X = x;
                                player.Position.Y = y;
                            }
							
							// badelineVanish badelineArg
                            if (nextCommand.StartsWith("badelineVanish"))
                            {
                                string badelineVarName = ParseArg(args, 1, "");
                                BadelineDummy badeline; //= Vars[badelineVarName] as BadelineDummy;
                                if (!Vars.ContainsKey(badelineVarName))
                                {
                                    badeline = null;
                                }
                                else
                                {
                                    badeline = Vars[badelineVarName] as BadelineDummy;
                                }
                                if (badeline != null)
                                    badeline.Vanish();
                                Vars.Remove(badelineVarName);
                            }

                            // crumbleWallsOnRumble
                            if (nextCommand.StartsWith("crumbleWallsOnRumble"))
                            {
                                foreach (Entity entity in (cutsceneTrigger.Scene as Level).Tracker.GetEntities<CrumbleWallOnRumble>())
                                {
                                    CrumbleWallOnRumble crumbleWallOnRumble = (CrumbleWallOnRumble)entity;
                                    crumbleWallOnRumble.Break();
                                }
                            }

                            // cameraZoom zoom
                            if (nextCommand.StartsWith("cameraZoom"))
                            {

                                float zoom = ParseArg(args, 1, 0);
                                (cutsceneTrigger.Scene as Level).Camera.Zoom = zoom;
                            }

                            // NEW in 1.1
                            // spawnBadelineChaser x y relativeToPlayer? canChangeMusic? permament?
                            if (nextCommand.StartsWith("spawnBadelineChaser"))
                            {

                                float x = ParseArg(args, 1, 0);
                                float y = ParseArg(args, 2, 0);
                                bool relativeToPlayer = ParseArg(args, 3, false);
                                bool canChangeMusic = ParseArg(args, 4, false);
                                bool permament = ParseArg(args, 5, false);
                                //foreach (BadelineOldsite badeline2 in Scene.Tracker.GetEntities<BadelineOldsite>())
                                // {
                                badelineChaserIndex++;
                                //}
                                Vector2 pos = new Vector2(x, y);
                                if (relativeToPlayer)
                                {
                                    pos += player.Position;
                                } else
                                {
                                    pos += (cutsceneTrigger.Scene as Level).LevelOffset;
                                }
                                
                                BadelineOldsite badeline;
                                (cutsceneTrigger.Scene as Level).Add(badeline = new BadelineOldsite(pos, badelineChaserIndex));
                                DynData<BadelineOldsite> data = new DynData<BadelineOldsite>(badeline);
                                data.Set("canChangeMusic", canChangeMusic);
                                if (permament)
                                {
                                    storedChaserCount++;
                                    //storedChasersPos.SetValue(pos, storedChaserCount - 1);
                                    Array.Resize<Vector2>(ref storedChasersPos, storedChasersPos.Length + 1);
                                    storedChasersPos[storedChasersPos.Length - 1] = pos;
                                    if (canChangeMusic)
                                    {
                                        storedChasersCanChangeMusic = true;
                                    }
                                }
                            }

                            // NEW in 1.1
                            // badelineChasersStopChasing threaded?
                            if (nextCommand.StartsWith("badelineChasersStopChasing"))
                            {
                                bool threaded = true;
                                badelineChaserIndex = 0;
                                foreach (BadelineOldsite badeline in (cutsceneTrigger.Scene as Level).Tracker.GetEntities<BadelineOldsite>())
                                {
                                    if (threaded)
                                    {
                                        cutsceneTrigger.Add(new Coroutine(badeline_StopChasing(badeline), true));
                                    }

                                }
                            }

                            // NEW in 1.1
                            // playMusic music progress
                            if (nextCommand.StartsWith("playMusic"))
                            {
                                string music = ParseArg(args, 1, "");
                                // backwards compat with 1.1
                                if (music.StartsWith("event:"))
                                {
                                    Audio.Play(music);
                                    
                                } else
                                {
                                    int Progress = (int)ParseArg(args, 2, 0);
                                    Session session = (cutsceneTrigger.Scene as Level).Session;
                                    session.Audio.Music.Event = SFX.EventnameByHandle(music);
                                    bool flag = Progress != 0;
                                    if (flag)
                                    {
                                        session.Audio.Music.Progress = Progress;
                                    }
                                    session.Audio.Apply(false);
                                }
                                

                                //Audio.Play(music);
                            }

                            #region flowControl
                            if (nextCommand.StartsWith("exec "))
                            {
                                string calledFunction = ParseArg(args, 1, "");
                                cutsceneTrigger.Add(new Coroutine(InterpretScripts(calledFunction, false)));
                            }

                            // when skipped, this immediately interprets everything
                            // execWait functionName
                            if (nextCommand.StartsWith("execWait"))
                            {
                                string calledFunction = ParseArg(args, 1, "");
                                Logger.Log("Cutscene Helper", "Calling a function:");
                                InterpretSkipped(calledFunction, false);
                            }

                            // label Name
                            // No actual code to this

                            // goto labelName
                            if (nextCommand.StartsWith("goto"))
                            {
                                string labelName = ParseArg(args, 1, "");
                                if (labelName != "")
                                {
                                    int labelIndex = SeekLabel(labelName, lines);
                                    if (labelIndex > -1)
                                    {
                                        //reader.BaseStream.Seek(labelIndex, SeekOrigin.Begin);
                                        //reader.DiscardBufferedData();
                                        //reader.ReadLine();
                                        currentIndex = labelIndex;
                                    }

                                }

                            }

                            // if varName condition value
                            if (args[0] == "if")
                            {
                                string varName = ParseArg(args, 1, "");
                                string varValue = Vars[varName].ToString();
                                string condition = ParseArg(args, 2, "=");
                                string targetValue = ParseArg(args, 3, "");
                                if (condition == "=")
                                {
                                    if (varValue == targetValue)
                                    {
                                        OnTrue();
                                    }
                                    else
                                    {
                                        OnFalse();
                                    }
                                }
                                if (condition == ">")
                                {
                                    if (float.Parse(varValue) > float.Parse(targetValue))
                                    {
                                        OnTrue();
                                    }
                                    else
                                    {
                                        OnFalse();
                                    }
                                }
                                if (condition == "<")
                                {
                                    if (float.Parse(varValue) < float.Parse(targetValue))
                                    {
                                        OnTrue();
                                    }
                                    else
                                    {
                                        OnFalse();
                                    }
                                }
                                if (condition == ">=")
                                {
                                    if (float.Parse(varValue) >= float.Parse(targetValue))
                                    {
                                        OnTrue();
                                    }
                                    else
                                    {
                                        OnFalse();
                                    }
                                }
                                if (condition == "<=")
                                {
                                    if (float.Parse(varValue) <= float.Parse(targetValue))
                                    {
                                        OnTrue();
                                    }
                                    else
                                    {
                                        OnFalse();
                                    }
                                }

                                void OnTrue()
                                {

                                }

                                void OnFalse()
                                {
                                    int iflevel = 1;
                                    for (; ; )
                                    {
                                        currentIndex++;
                                        string next = lines[currentIndex].TrimStart(' ');
                                        if (next.StartsWith("endIf"))
                                        {
                                            iflevel--;
                                            if (iflevel < 1)
                                                break;
                                        }
                                        if (next.StartsWith("if"))
                                        {
                                            iflevel++;
                                        }
                                    }
                                }
                            }

                            #endregion

                            #region Variables
                            // setVar varName value
                            if (nextCommand.StartsWith("setVar"))
                            {
                                string varName = ParseArg(args, 1, "");
                                string value = ParseArg(args, 2, "");
                                if (!Vars.ContainsKey(varName))
                                {
                                    Vars.Add(varName, value);
                                }
                                else
                                {
                                    Vars[varName] = value;
                                }

                            }

                            // addVarNum varName value
                            if (nextCommand.StartsWith("addVarNum"))
                            {
                                string varName = ParseArg(args, 1, "");
                                float value = ParseArg(args, 2, 0f);
                                if (Vars.ContainsKey(varName))
                                {
                                    Vars[varName] = float.Parse(Vars[varName].ToString()) + value;
                                }
                            }

                            // getCassetteBlockIndex retVar
                            if (nextCommand.StartsWith("getCassetteBlockIndex"))
                            {
                                string varName = ParseArg(args, 1, "");
                                if (!Vars.ContainsKey(varName))
                                {
                                    Vars.Add(varName, CSHelperModule.casetteIndex);
                                }
                                else
                                {
                                    Vars[varName] = CSHelperModule.casetteIndex;
                                }
                            }

                            // getCoreMode retVar
                            if (nextCommand.StartsWith("getCoreMode"))
                            {
                                string varName = ParseArg(args, 1, "");
                                string value = (cutsceneTrigger.Scene as Level).CoreMode == Session.CoreModes.Hot ? "hot" : "cold";
                                if (!Vars.ContainsKey(varName))
                                {
                                    Vars.Add(varName, value);
                                }
                                else
                                {
                                    Vars[varName] = value;
                                }
                            }

                            #endregion

                            #region Debug
                            // debugLog text
                            if (nextCommand.StartsWith("debugLog"))
                            {
                                Logger.Log("CutsceneHelper", nextCommand.Substring("debugLog ".Length));
                            }
                            #endregion

                            // setPlayerDashes count
                            if (nextCommand.StartsWith("setPlayerDashes"))
                            {
                                int dashes = (int)ParseArg(args, 1, 0);
                                player.Dashes = dashes;
                                player.UpdateHair(true);
                            }

                            // setPlayerMaxDashes count
                            if (nextCommand.StartsWith("setPlayerMaxDashes"))
                            {
                                int dashes = (int)ParseArg(args, 1, 0);
                                (cutsceneTrigger.Scene as Level).Session.Inventory.Dashes = dashes;
                                if (player.Dashes > dashes)
                                {
                                    player.Dashes = dashes;
                                }
                                player.UpdateHair(true);
                            }

                            // playerRefilOnGround true?
                            if (nextCommand.StartsWith("playerRefilOnGround") || nextCommand.StartsWith("playerRefillOnGround"))
                            {
                                bool playerRefilOnGround = ParseArg(args, 1, true);
                                (cutsceneTrigger.Scene as Level).Session.Inventory.NoRefills = playerRefilOnGround;
                            }

                            #region Misc
                            // setCassetteBlockTempo tempo
                            if (nextCommand.StartsWith("setCassetteBlockTempo"))
                            {
                                float Tempo = ParseArg(args, 1, 1f);
                                SetCassetteTempo(Tempo);
                            }

                            // activateDreamBlocks threaded?
                            if (args[0] == "activateDreamBlocks")
                            {
                                level.Session.Inventory.DreamDash = true;
                                foreach (DreamBlock dreamBlock in cutsceneTrigger.Scene.Tracker.GetEntities<DreamBlock>())
                                {
                                    dreamBlock.ActivateNoRoutine();

                                }
                            }

                            ExecSharedFuncs(args);
                            #endregion
                        }

                    }
                }
            }
        }
    }
}
