﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Celeste;
using Microsoft.Xna.Framework;
using Monocle;

namespace CutsceneHelper
{
    public class CutsceneTrigger : Trigger
    {
        public string cutsceneID;
        private EntityID id;
        bool endLevel;
        bool onlyOnce;
        bool lockMovement;
        public bool infinite;
        public bool triggered = false;

        public CutsceneTrigger(EntityData data, Vector2 offset, EntityID entId) : base(data, offset)
        {
            cutsceneID = data.Attr("cutsceneFileName", "");
            endLevel = data.Bool("endLevel", false);
            onlyOnce = data.Bool("oneTime", true);
            infinite = data.Bool("infiniteRetrigger", false);
            lockMovement = data.Bool("lockMovement", true);
            id = entId;
        }

        public override void OnEnter(Player player)
        {
            if (!triggered)
            {
                triggered = true;
                if ((Scene as Level).Session.GetFlag("DoNotLoad" + id))
                {
                    return;
                }
                int state = player.StateMachine.State;
                if (lockMovement)
                {
                    Cutscene cutscene = new Cutscene(cutsceneID, endLevel, lockMovement, this);
                    Scene.Add(cutscene);
                    cutscene.cutsceneTrigger = this;
                } else
                {
                    CutsceneInterpreter interpreter = new CutsceneInterpreter(cutsceneID, endLevel, lockMovement);
                    interpreter.cutsceneTrigger = this;
                    Add(new Coroutine(interpreter.InterpretScripts()));
                }
                if (onlyOnce)
                    (Scene as Level).Session.SetFlag("DoNotLoad" + id, true); // Sets flag to not load
            }
            
        }
    }
}
